#include "Shape.h"
#include <iostream>

#define EIGEN_DONT_ALIGN_STATICALLY
#include <Eigen/Dense>
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include "GLSL.h"
#include "Program.h"
#include "MatrixStack.h"

#define TINYOBJLOADER_IMPLEMENTATION
#include "tiny_obj_loader.h"

using namespace std;
using namespace Eigen;

Shape::Shape() :
	eleBufID(0),
	posBufID(0),
	norBufID(0),
	texBufID(0), 
   vaoID(0)
{
}

Shape::~Shape()
{
}

Vector3f CrossProduct(Vector3f lhs, Vector3f rhs) {
	Vector3f normal = Vector3f(0, 0, 0);
	normal(0) = lhs(1) * rhs(2) - lhs(2) * rhs(1);
	normal(1) = lhs(2) * rhs(0) - lhs(0) * rhs(2);
	normal(2) = lhs(0) * rhs(1) - lhs(1) * rhs(0);
	return normal;
}

void Shape::loadMesh(const string &meshName)
{
	// Load geometry
	// Some obj files contain material information.
	// We'll ignore them for this assignment.
	vector<tinyobj::shape_t> shapes;
	vector<tinyobj::material_t> objMaterials;
	string errStr;
	bool rc = tinyobj::LoadObj(shapes, objMaterials, errStr, meshName.c_str());
	float *calcNorms;

	if(!rc) {
		cerr << errStr << endl;
	} else {
		calcNorms = (float *)calloc(sizeof(float), shapes[0].mesh.positions.size());

		for (unsigned int i = 0; i * 3 < shapes[0].mesh.indices.size(); i++) {
			Vector3f calcNorm = CrossProduct(Vector3f(shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 1] * 3 + 0] - shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 0] * 3 + 0],
													  shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 1] * 3 + 1] - shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 0] * 3 + 1],
													  shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 1] * 3 + 2] - shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 0] * 3 + 2]),
											 Vector3f(shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 2] * 3 + 0] - shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 0] * 3 + 0],
													  shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 2] * 3 + 1] - shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 0] * 3 + 1],
													  shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 2] * 3 + 2] - shapes[0].mesh.positions[shapes[0].mesh.indices[i * 3 + 0] * 3 + 2]));
			calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 0] += calcNorm(0);
			calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 1] += calcNorm(1);
			calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 2] += calcNorm(2);
			calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 0] += calcNorm(0);
			calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 1] += calcNorm(1);
			calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 2] += calcNorm(2);
			calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 0] += calcNorm(0);
			calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 1] += calcNorm(1);
			calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 2] += calcNorm(2);
		}

		for (unsigned int i = 0; i * 3 < shapes[0].mesh.indices.size(); i++) {
			Vector3f tempNorm = Vector3f(calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 0],
										 calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 1],
										 calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 2]);
			tempNorm.normalize();
			calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 0] = tempNorm(0);
			calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 1] = tempNorm(1);
			calcNorms[shapes[0].mesh.indices[i * 3 + 0] * 3 + 2] = tempNorm(2);
			
			tempNorm = Vector3f(calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 0],
								calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 1],
								calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 2]);
			tempNorm.normalize();
			calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 0] = tempNorm(0);
			calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 1] = tempNorm(1);
			calcNorms[shapes[0].mesh.indices[i * 3 + 1] * 3 + 2] = tempNorm(2);
			
			tempNorm = Vector3f(calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 0],
								calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 1],
								calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 2]);
			tempNorm.normalize();
			calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 0] = tempNorm(0);
			calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 1] = tempNorm(1);
			calcNorms[shapes[0].mesh.indices[i * 3 + 2] * 3 + 2] = tempNorm(2);
		}

		std::vector<float> newNorms = std::vector<float>(shapes[0].mesh.positions.size());
		for (int i = 0; i < shapes[0].mesh.positions.size(); i++) {
			newNorms[i] = calcNorms[i];
		}
		posBuf = shapes[0].mesh.positions;
		norBuf = shapes[0].mesh.normals;
		texBuf = shapes[0].mesh.texcoords;
		eleBuf = shapes[0].mesh.indices;
	}
}

void Shape::resize() {
  float minX, minY, minZ;
   float maxX, maxY, maxZ;
   float scaleX, scaleY, scaleZ;
   float shiftX, shiftY, shiftZ;
   float epsilon = 0.001f;

   minX = minY = minZ = 1.1754E+38F;
   maxX = maxY = maxZ = -1.1754E+38F;

   //Go through all vertices to determine min and max of each dimension
   for (size_t v = 0; v < posBuf.size() / 3; v++) {
		if(posBuf[3*v+0] < minX) minX = posBuf[3*v+0];
		if(posBuf[3*v+0] > maxX) maxX = posBuf[3*v+0];

		if(posBuf[3*v+1] < minY) minY = posBuf[3*v+1];
		if(posBuf[3*v+1] > maxY) maxY = posBuf[3*v+1];

		if(posBuf[3*v+2] < minZ) minZ = posBuf[3*v+2];
		if(posBuf[3*v+2] > maxZ) maxZ = posBuf[3*v+2];
	}

	//From min and max compute necessary scale and shift for each dimension
   float maxExtent, xExtent, yExtent, zExtent;
   xExtent = maxX-minX;
   yExtent = maxY-minY;
   zExtent = maxZ-minZ;
   if (xExtent >= yExtent && xExtent >= zExtent) {
      maxExtent = xExtent;
   }
   if (yExtent >= xExtent && yExtent >= zExtent) {
      maxExtent = yExtent;
   }
   if (zExtent >= xExtent && zExtent >= yExtent) {
      maxExtent = zExtent;
   }
   scaleX = 2.0f /maxExtent;
   shiftX = minX + (xExtent/ 2.0f);
   scaleY = 2.0f / maxExtent;
   shiftY = minY + (yExtent / 2.0f);
   scaleZ = 2.0f/ maxExtent;
   shiftZ = minZ + (zExtent)/2.0f;

   //Go through all verticies shift and scale them
	for (size_t v = 0; v < posBuf.size() / 3; v++) {
		posBuf[3*v+0] = (posBuf[3*v+0] - shiftX) * scaleX;
		assert(posBuf[3*v+0] >= -1.0 - epsilon);
		assert(posBuf[3*v+0] <= 1.0 + epsilon);
		posBuf[3*v+1] = (posBuf[3*v+1] - shiftY) * scaleY;
		assert(posBuf[3*v+1] >= -1.0 - epsilon);
		assert(posBuf[3*v+1] <= 1.0 + epsilon);
		posBuf[3*v+2] = (posBuf[3*v+2] - shiftZ) * scaleZ;
		assert(posBuf[3*v+2] >= -1.0 - epsilon);
		assert(posBuf[3*v+2] <= 1.0 + epsilon);
	}
}

void Shape::init()
{
   // Initialize the vertex array object
   glGenVertexArrays(1, &vaoID);
   glBindVertexArray(vaoID);

	// Send the position array to the GPU
	glGenBuffers(1, &posBufID);
	glBindBuffer(GL_ARRAY_BUFFER, posBufID);
	glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_STATIC_DRAW);
	
	// Send the normal array to the GPU
	if(norBuf.empty()) {
		norBufID = 0;
	} else {
		glGenBuffers(1, &norBufID);
		glBindBuffer(GL_ARRAY_BUFFER, norBufID);
		glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_STATIC_DRAW);
	}
	
	// Send the texture array to the GPU
	if(texBuf.empty()) {
		texBufID = 0;
	} else {
		glGenBuffers(1, &texBufID);
		glBindBuffer(GL_ARRAY_BUFFER, texBufID);
		glBufferData(GL_ARRAY_BUFFER, texBuf.size()*sizeof(float), &texBuf[0], GL_STATIC_DRAW);
	}
	
	// Send the element array to the GPU
	glGenBuffers(1, &eleBufID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eleBufID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, eleBuf.size()*sizeof(unsigned int), &eleBuf[0], GL_STATIC_DRAW);
	
	// Unbind the arrays
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	assert(glGetError() == GL_NO_ERROR);
}

void Shape::draw(const shared_ptr<Program> prog) const
{
	int h_pos, h_nor, h_tex;
	h_pos = h_nor = h_tex = -1;

   glBindVertexArray(vaoID);
	// Bind position buffer
	h_pos = prog->getAttribute("vertPos");
	GLSL::enableVertexAttribArray(h_pos);
	glBindBuffer(GL_ARRAY_BUFFER, posBufID);
	glVertexAttribPointer(h_pos, 3, GL_FLOAT, GL_FALSE, 0, (const void *)0);
	
	// Bind normal buffer
	h_nor = prog->getAttribute("vertNor");
	if(h_nor != -1 && norBufID != 0) {
		GLSL::enableVertexAttribArray(h_nor);
		glBindBuffer(GL_ARRAY_BUFFER, norBufID);
		glVertexAttribPointer(h_nor, 3, GL_FLOAT, GL_FALSE, 0, (const void *)0);
	}

	if (texBufID != 0) {	
		// Bind texcoords buffer
		h_tex = prog->getAttribute("vertTex");
		if(h_tex != -1 && texBufID != 0) {
			GLSL::enableVertexAttribArray(h_tex);
			glBindBuffer(GL_ARRAY_BUFFER, texBufID);
			glVertexAttribPointer(h_tex, 2, GL_FLOAT, GL_FALSE, 0, (const void *)0);
		}
	}
	
	// Bind element buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eleBufID);
	
	// Draw
	glDrawElements(GL_TRIANGLES, (int)eleBuf.size(), GL_UNSIGNED_INT, (const void *)0);
	
	// Disable and unbind
	if(h_tex != -1) {
		GLSL::disableVertexAttribArray(h_tex);
	}
	if(h_nor != -1) {
		GLSL::disableVertexAttribArray(h_nor);
	}
	GLSL::disableVertexAttribArray(h_pos);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
