/* Lab 5 base code - transforms using local matrix functions 
   to be written by students - 
	CPE 471 Cal Poly Z. Wood + S. Sueda
*/
#include <iostream>
#define GLEW_STATIC
#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <time.h>
#include <stdlib.h>

#include "GLSL.h"
#include "Program.h"
#include "MatrixStack.h"
#include "Shape.h"
#define LIGHT 0x01
#define SPHERE 0x02
#define CUBE 0x04
#define NONE 0x10
#define TRACED 0x20
#define DYING 0x40
#define DEAD 0x80

using namespace std;
using namespace Eigen;

GLFWwindow *window; // Main application window
string RESOURCE_DIR = "../resources/"; // Where the resources are loaded from
auto P = make_shared<MatrixStack>();
auto V = make_shared<MatrixStack>();
shared_ptr<Program> prog;
shared_ptr<Program> pnt;
shared_ptr<Shape> bunny;
shared_ptr<Shape> dog;
shared_ptr<Shape> sphere;
shared_ptr<Shape> cube;
shared_ptr<Shape> nef;
Vector4f pntLight;
Vector3f playerPos;
Vector3f lightColor;
Vector3f camRotation;
Vector3f tracePosition;

GLuint VertexArrayID;
static GLfloat g_vertex_buffer_data[] = {0, 0, 0};
static GLfloat g_color_buffer_data[] = {1, 1, 1};
GLuint vertexbuffer;
GLuint colorbuffer;

int g_width, g_height, mat = 0, timer, score;
float aspect, moveX = 0, moveY = 0, disNorm, camX = M_PI/-2, camY = 0,  yMinMax = 80 * M_PI / 180, speed = 1, sunRot = 45;
double mPosX = 0, mPosY = 0;
bool wDown, sDown, aDown, dDown, spaceDown, shiftDown, leftDown, rightDown, upDown, downDown, mouseDown, mousePressed, wave1, wave2, cleanDead;

Vector3f vec3(float x, float y, float z) {
	return Vector3f(x, y, z);
}

Vector3f vec3(float s) {
	return Vector3f(s, s, s);
}

typedef struct {
	Vector3f rotate;
	float amplitude;
	float offset;
	float deltaTime;
	float timer;
	bool extend;
	Vector3f (*animate)(Vector3f, float, float, float, float*, bool*);
} Animator;

typedef struct {
   Vector3f amb;
   Vector3f dif;
   Vector3f spc;
   float shine;
} Material;

Material GetMaterial(int i) {
   Material mat;
   switch (i) {
   case 0:
      mat.amb = vec3(0.02f, 0.2f, 0.04f);
      mat.dif = vec3(0.00f, 0.9f, 0.16f);
      mat.spc = vec3(0.14f, 0.8f, 0.20f);
      mat.shine = 120.0f;
      break;
   case 1:
      mat.amb = vec3(0.1f, 0.1f, 0.11f);
      mat.dif = vec3(0.6f, 0.6f, 0.75f);
      mat.spc = vec3(0.4f, 0.4f, 0.5f);
      mat.shine = 100.0f;
      break;
   case 2:
      mat.amb = vec3(0.13f, 0.12f, 0.03f);
      mat.dif = vec3(0.78f, 0.57f, 0.01f);
      mat.spc = vec3(0.99f, 0.94f, 0.71f);
      mat.shine = 150.0f;
      break;
   case 3:
      mat.amb = vec3(0.29f, 0.07f, 0.02f);
      mat.dif = vec3(0.61f, 0.17f, 0.08f);
      mat.spc = vec3(0.56f, 0.14f, 0.09f);
      mat.shine = 15.0f;
      break;
   default: 
      mat.amb = vec3(0, 0, 0);
      mat.dif = vec3(0, 0, 0);
      mat.spc = vec3(0, 0, 0);
      mat.shine = 1;
   }
   return mat;
}

Material CombineMaterials(Material mat1, Material mat2, float dis) {
   Material matN;
   matN.amb = mat1.amb * dis + mat2.amb * (1-dis);
   matN.amb = mat1.dif * dis + mat2.dif * (1-dis);
   matN.amb = mat1.spc * dis + mat2.spc * (1-dis);
   matN.shine = mat1.shine * dis + mat2.shine * (1-dis);
   return matN;
}

typedef struct {
	bool on;
	Vector3f velocity;
} Rigidbody;

typedef struct GameObject {
	Vector3f position;
	Vector3f rotation;
	Vector3f joint;
	Vector3f scale;
	shared_ptr<Program> prog;
	shared_ptr<Shape> shape;
	int numAnimators;
	Animator *animators;
	Material mat;
	Rigidbody rig;
	int numChildren;
	GameObject *children;
	int flag;
} GameObject;

float vecDistance(Vector3f p1, Vector3f p2) {
   return sqrt((p1(0)-p2(0))*(p1(0)-p2(0)) + (p1(1)-p2(1))*(p1(1)-p2(1)) + (p1(2)-p2(2))*(p1(2)-p2(2)));
}

GameObject createGameObject(Vector3f pos, Vector3f rot, Vector3f jnt, Vector3f scl, Vector3f vel, shared_ptr<Program> prog, shared_ptr<Shape> shape, int sMat, int flag) {
	GameObject obj;
	obj.position = pos;
	obj.rotation = rot;
	obj.joint = jnt;
	obj.scale = scl;
	obj.prog = prog;
	obj.shape = shape;
	obj.numAnimators = 0;
	obj.animators = 0;
	obj.mat = GetMaterial(sMat);
	obj.rig.on = false;
	obj.rig.velocity = vel;
	obj.numChildren = 0;
	obj.children = 0;
	obj.flag = flag;
	return obj;
}

Animator createAnimator(Vector3f rot, float ampl, float off, float del, Vector3f (*callback)(Vector3f, float, float, float, float*, bool*)) {
	Animator animator;
	animator.rotate = rot;
	animator.amplitude = ampl;
	animator.offset = off;
	animator.deltaTime = del;
	animator.timer = 0;
	animator.extend = false;
	animator.animate = callback;
	return animator;
}

static GameObject addChild(GameObject parent, GameObject child) {
	if (parent.numChildren) {
		parent.children = (GameObject *)realloc(parent.children, (parent.numChildren + 1) * sizeof(GameObject));
		memcpy(parent.children + parent.numChildren, &child, sizeof(GameObject));
	}
	else {
		parent.children = (GameObject *)malloc(sizeof(GameObject));
		memcpy(parent.children, &child, sizeof(GameObject));
	}

	parent.numChildren++;

	return parent;
}

GameObject addAnimator(GameObject gameObject, Animator animator) {
	if (gameObject.numAnimators) {
		gameObject.animators = (Animator *)realloc(gameObject.animators, (gameObject.numAnimators + 1) * sizeof(Animator));
		memcpy(gameObject.animators + gameObject.numAnimators, &animator, sizeof(Animator));
	}
	else {
		gameObject.animators = (Animator *)malloc(sizeof(Animator));
		memcpy(gameObject.animators, &animator, sizeof(Animator));
	}

	gameObject.numAnimators++;

	return gameObject;
}

Vector3f playAnimation(Animator *animator) {
	return animator->animate(animator->rotate, animator->amplitude, animator->offset, animator->deltaTime, &animator->timer, &animator->extend);
}

//helper function to set materials
void SetMaterial(Material mat, shared_ptr<Program> prog) {
	glUniform3f(prog->getUniform("MatAmb"), mat.amb(0), mat.amb(1), mat.amb(2));
	glUniform3f(prog->getUniform("MatDif"), mat.dif(0), mat.dif(1), mat.dif(2));
	glUniform3f(prog->getUniform("MatSpec"), mat.spc(0), mat.spc(1), mat.spc(2));
	glUniform1f(prog->getUniform("MatShine"), mat.shine);
}

static void error_callback(int error, const char *description) {
	cerr << description << endl;
}

static void key_callback(GLFWwindow *window, int key, int scancode, int action, int mods) {
	if(key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
		glfwSetWindowShouldClose(window, GL_TRUE);
	}
	if (key == GLFW_KEY_X && action == GLFW_PRESS) {
		wave1 = !wave1;
		wave2 = false;
	}
	if (key == GLFW_KEY_C && action == GLFW_PRESS) {
		wave2 = !wave2;
		wave1 = false;
	}
	if (key == GLFW_KEY_W && action == GLFW_PRESS) {
		wDown = true;
	}
	if (key == GLFW_KEY_W && action == GLFW_RELEASE) {
		wDown = false;
	}
	if (key == GLFW_KEY_S && action == GLFW_PRESS) {
		sDown = true;
	}
	if (key == GLFW_KEY_S && action == GLFW_RELEASE) {
		sDown = false;
	}
	if (key == GLFW_KEY_A && action == GLFW_PRESS) {
		aDown = true;
	}
	if (key == GLFW_KEY_A && action == GLFW_RELEASE) {
		aDown = false;
	}
	if (key == GLFW_KEY_D && action == GLFW_PRESS) {
		dDown = true;
	}
	if (key == GLFW_KEY_D && action == GLFW_RELEASE) {
		dDown = false;
	}
	if (key == GLFW_KEY_SPACE && action == GLFW_PRESS) {
		spaceDown = true;
	}
	if (key == GLFW_KEY_SPACE && action == GLFW_RELEASE) {
		spaceDown = false;
	}
	if (key == GLFW_KEY_LEFT_SHIFT && action == GLFW_PRESS) {
		shiftDown = true;
	}
	if (key == GLFW_KEY_LEFT_SHIFT && action == GLFW_RELEASE) {
		shiftDown = false;
	}
	if (key == GLFW_KEY_LEFT && action == GLFW_PRESS) {
		leftDown = true;
	}
	if (key == GLFW_KEY_LEFT && action == GLFW_RELEASE) {
		leftDown = false;
	}
	if (key == GLFW_KEY_RIGHT && action == GLFW_PRESS) {
		rightDown = true;
	}
	if (key == GLFW_KEY_RIGHT && action == GLFW_RELEASE) {
		rightDown = false;
	}
	if (key == GLFW_KEY_Q && action == GLFW_PRESS) {
		leftDown = true;
	}
	if (key == GLFW_KEY_Q && action == GLFW_RELEASE) {
		leftDown = false;
	}
	if (key == GLFW_KEY_E && action == GLFW_PRESS) {
		rightDown = true;
	}
	if (key == GLFW_KEY_E && action == GLFW_RELEASE) {
		rightDown = false;
	}
	if (key == GLFW_KEY_UP && action == GLFW_PRESS) {
		upDown = true;
	}
	if (key == GLFW_KEY_UP && action == GLFW_RELEASE) {
		upDown = false;
	}
	if (key == GLFW_KEY_DOWN && action == GLFW_PRESS) {
		downDown = true;
	}
	if (key == GLFW_KEY_DOWN && action == GLFW_RELEASE) {
		downDown = false;
	}
}


static void mouse_callback(GLFWwindow *window, int button, int action, int mods) {
   if (button == 0 && action == GLFW_PRESS) mouseDown = true;
   if (button == 0 && action == GLFW_RELEASE) mouseDown = false;
   if (button == 1 && action == GLFW_PRESS) mousePressed = true;
}

static void resize_callback(GLFWwindow *window, int width, int height) {
   g_width = width;
   g_height = height;
   glViewport(0, 0, width, height);
}

/*Note that any gl calls must always happen after a GL state is initialized */
static void initGeom() {

	//generate the VAO
	glGenVertexArrays(1, &VertexArrayID);
	glBindVertexArray(VertexArrayID);

	//generate vertex buffer to hand off to OGL
	glGenBuffers(1, &vertexbuffer);
	//set the current state to focus on our vertex buffer
	glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
	//actually memcopy the data - only do this once
	glBufferData(GL_ARRAY_BUFFER, sizeof(g_vertex_buffer_data), g_vertex_buffer_data, GL_STATIC_DRAW);

	glGenBuffers(1, &colorbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(g_color_buffer_data), g_color_buffer_data, GL_STATIC_DRAW);
}

static void init() {
	GLSL::checkVersion();

	timer = 0;
	score = 0;
	lightColor = vec3(1);
	playerPos = vec3(0);
	// Set background color.
	glClearColor(.12f, .34f, .56f, 1.0f);
	// Enable z-buffer test.
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Initialize mesh.
	bunny = make_shared<Shape>();
	bunny->loadMesh(RESOURCE_DIR + "bunny.obj");
	bunny->resize();
	bunny->init();
	sphere = make_shared<Shape>();
	sphere->loadMesh(RESOURCE_DIR + "sphere.obj");
	sphere->resize();
	sphere->init();
	cube = make_shared<Shape>();
	cube->loadMesh(RESOURCE_DIR + "cube.obj");
	cube->resize();
	cube->init();
	nef = make_shared<Shape>();
	nef->loadMesh(RESOURCE_DIR + "Nefertiti.obj");
	nef->resize();
	nef->init();
	dog = make_shared<Shape>();
	dog->loadMesh(RESOURCE_DIR + "dog.obj");
	dog->resize();
	dog->init();

	// Initialize the GLSL program.
	prog = make_shared<Program>();
	prog->setVerbose(true);
	prog->setShaderNames(RESOURCE_DIR + "vert_phong.glsl", RESOURCE_DIR + "frag_phong.glsl");
	prog->init();
	prog->addUniform("P");
	prog->addUniform("M");
	prog->addUniform("V");
	prog->addUniform("light");
	prog->addUniform("lightColor");
	prog->addUniform("MatAmb");
	prog->addUniform("MatDif");
	prog->addUniform("MatSpec");
	prog->addUniform("MatShine");
	prog->addUniform("disNorm");
	prog->addAttribute("vertPos");
	prog->addAttribute("vertNor");
	prog->addAttribute("vertTex");

	// Initialize the GLSL program 2.
	pnt = make_shared<Program>();
	pnt->setVerbose(true);
	pnt->setShaderNames(RESOURCE_DIR + "point_vert.glsl", RESOURCE_DIR + "point_frag.glsl");
	pnt->init();
	pnt->addUniform("P");
	pnt->addUniform("MV");
	pnt->addUniform("aspect");
	pnt->addAttribute("vertPos");
	pnt->addAttribute("vertexColor");
	
	initGeom();
}

static GameObject createPrefab1(Vector3f pos, float radius) {
	GameObject newScene = createGameObject(pos, vec3(0), vec3(0), vec3(1), vec3(0), prog, cube, 0, SPHERE | NONE);
	int i, j, k;
	float epsilon = 1.25f;
	
	for (i = -radius; i < radius; i++) {
	   for (j = -radius; j < radius; j++) {
	      for (k = -radius; k < radius; k++) {
	         float dis = sqrt(i*i + j*j + k*k);
	         if (radius - dis < epsilon && radius - dis > 0)
	            newScene = addChild(newScene, createGameObject(vec3(i, j, k), vec3(0), vec3(0), vec3(1), vec3(0), prog, cube, 1, CUBE));
	      }
	   }
	}

   return newScene;
}

static GameObject setupScene() {
	GameObject newScene = createGameObject(vec3(0, 0, 0), vec3(0), vec3(0), vec3(1), vec3(0), prog, cube, 0, NONE);
	GameObject light = createGameObject(vec3(0), vec3(0, 0, 45), vec3(50, 0, 0), vec3(1), vec3(0), prog, sphere, 2, LIGHT | NONE);
	GameObject floor = createGameObject(vec3(0, -12.5f, 0), vec3(0), vec3(0), vec3(500, .1f, 500), vec3(0), prog, cube, 0, 0);
   srand(time(NULL));

	newScene = addChild(newScene, light);
	newScene = addChild(newScene, floor);
	newScene = addChild(newScene, createPrefab1(vec3(0, 25, -75), 6));
	
	
	return newScene;
}

Vector3f explosionAtPoint(Vector3f origin, Vector3f direction, float force, float radius) {
   float distance = vecDistance(origin, direction);
   return direction * (distance > radius ? 0 : (radius - distance) * force);
}

Vector3f traceLine(GameObject *scene, float minDistance, float maxDistance, float disDelta) {
   Vector3f line = camRotation;
   line.normalize();
   float d, min = maxDistance, max = minDistance;
   
   for (int i = 0; i < scene->numChildren; i++) {
      if (scene->children[i].flag & SPHERE && !(scene->children[i].flag & DEAD)) {
         if (vecDistance(scene->children[i].position, playerPos) < min) min = vecDistance(scene->children[i].position, playerPos);
         if (vecDistance(scene->children[i].position, playerPos) > max) max = vecDistance(scene->children[i].position, playerPos);
      }
   }
   min -= 6;
   min = min < minDistance ? minDistance : min;
   max += 6;
   max = max > maxDistance ? maxDistance : max;
   
   for (d = min; d < max; d += disDelta) {
	   for (int i = 0; i < scene->numChildren; i++) {
	      if (scene->children[i].flag & SPHERE) {
	         for (int j = 0; j < scene->children[i].numChildren; j++) {
	            if (vecDistance(line * d + playerPos, scene->children[i].position + scene->children[i].children[j].position) <= 1) {
	               scene->children[i].flag |= TRACED;
	               return line * d + playerPos - scene->children[i].position;
	            }
	         }
	      }
	   }
   }
   return line * d;
}

static void drawGameObject(shared_ptr<MatrixStack> M, GameObject *gameObject) {
	M->pushMatrix();

	if (gameObject->prog) gameObject->prog->bind();

	Vector3f animateRotation = Vector3f(0, 0, 0);

	for (int a = 0; a < gameObject->numAnimators; a++) {
		animateRotation += playAnimation(gameObject->animators + a);
	}
	
	if (gameObject->flag & SPHERE) {
	   if (gameObject->numChildren < 50 && !(gameObject->flag & DYING)) {
	      
         for (int i = 0; i < gameObject->numChildren; i++) {
            gameObject->children[i].rig.velocity += explosionAtPoint(vec3(0), gameObject->children[i].position, .25f, 6);
            gameObject->children[i].rig.on = true;
         }
         gameObject->flag |= DYING;
      }
      else if (!gameObject->numChildren) {
         gameObject->flag |= DEAD;
      }
	}
	
   if (gameObject->rig.on) {
      gameObject->rig.velocity += vec3(0, -1.18f, 0)/30;
      gameObject->position += gameObject->rig.velocity;
      if (gameObject->position(1) <= -37) {
         gameObject->rig.on = false;
         gameObject->rig.velocity = vec3(0);
         gameObject->position = vec3(gameObject->position(0), -37, gameObject->position(2));
         gameObject->flag |= DYING;
      }
   }
   if (gameObject->flag & DYING) {
      gameObject->position -= vec3(0, .025f, 0);
      if (gameObject->position(1) < -39) {
         gameObject->flag |= DEAD;
         cleanDead = true;
      }
   }
   if ((gameObject->flag & TRACED) && (gameObject->flag & SPHERE)) {
      for (int i = 0; i < gameObject->numChildren; i++) {
         float distance = 7.5f;
         Vector3f curPos = gameObject->children[i].position;
         gameObject->children[i].rig.velocity += explosionAtPoint(tracePosition, curPos, .05f, distance);
         if (vecDistance(tracePosition, curPos) < distance) {
            if (!gameObject->children[i].rig.on) score++;
            gameObject->children[i].rig.on = true;
            gameObject->children[i].mat = CombineMaterials(GetMaterial(3), GetMaterial(1), vecDistance(tracePosition, curPos)/distance);
         }
      }
      printf("\rScore: %d", score);
      fflush(stdout);
      gameObject->flag &= ~TRACED;
	}

	M->translate(gameObject->position);
	M->rotate(gameObject->rotation(0) + animateRotation(0), Vector3f(1, 0, 0));
	M->rotate(gameObject->rotation(1) + animateRotation(1), Vector3f(0, 1, 0));
	M->rotate(gameObject->rotation(2) + animateRotation(2), Vector3f(0, 0, 1));
	M->translate(gameObject->joint);
	M->scale(gameObject->scale);
	if (!(gameObject->flag & NONE)) {
	   glUniformMatrix4fv(gameObject->prog->getUniform("P"), 1, GL_FALSE, P->topMatrix().data());
	   glUniformMatrix4fv(gameObject->prog->getUniform("M"), 1, GL_FALSE, M->topMatrix().data());
	   glUniformMatrix4fv(gameObject->prog->getUniform("V"), 1, GL_FALSE, V->topMatrix().data());
	   glUniform3f(gameObject->prog->getUniform("light"), pntLight(0), pntLight(1), pntLight(2));
	   glUniform3f(gameObject->prog->getUniform("lightColor"), lightColor(0), lightColor(1), lightColor(2));
	   glUniform1f(gameObject->prog->getUniform("disNorm"), disNorm);
	   SetMaterial(gameObject->mat, gameObject->prog);
	   gameObject->shape->draw(gameObject->prog);
	}
	
	if (gameObject->flag & LIGHT) pntLight = M->topMatrix() * Vector4f(0, 0, 0, 1);

	if (gameObject->prog) gameObject->prog->unbind();

	for (int i = 0; i < gameObject->numChildren; i++) {
		drawGameObject(M, &(gameObject->children[i]));
	}

	M->popMatrix();
}

static void drawPoints() {
	// Get current frame buffer size.
	int width, height;
	glfwGetFramebufferSize(window, &width, &height);
	float aspect = width / (float)height;
	glViewport(0, 0, width, height);

	// Clear framebuffer.
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);


	// Create the matrix stacks - please leave these alone for now
	auto PV = make_shared<MatrixStack>();
	auto MV = make_shared<MatrixStack>();
	// Apply orthographic projection.
	PV->pushMatrix();
	if (g_width > g_height) {
		PV->ortho(-1 * aspect, 1 * aspect, -1, 1, -1, 100.0f);
	}
	else {
		PV->ortho(-1, 1, -1 / aspect, 1 / aspect, -1, 100.0f);
	}
	MV->pushMatrix();
	MV->loadIdentity();
	
	// Draw the point using GLSL.
	pnt->bind();
	glUniformMatrix4fv(pnt->getUniform("P"), 1, GL_FALSE, P->topMatrix().data());
	glUniformMatrix4fv(pnt->getUniform("MV"), 1, GL_FALSE, MV->topMatrix().data());
	glUniform2f(pnt->getUniform("aspect"), width, height);

	//we need to set up how to pull vertices out of the vertex array
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
	//key function to get up how many elements to pull out at a time (3), (attribute, size, type, normalized, stride, array buffer offset)
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);

	// 2nd attribute buffer : colors
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);

	glPointSize(16);
	//actually draw from vertex 0, 3 vertices
	glDrawArrays(GL_POINTS, 0, 1);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);


	pnt->unbind();

	// Pop matrix stacks.
	glEnable(GL_DEPTH_TEST);
}

static void render(GameObject *scene) {
	// Get current frame buffer size.
	int width, height;
	glfwGetFramebufferSize(window, &width, &height);
	glViewport(0, 0, width, height);
	
	double mX = mPosX, mY = mPosY;
   glfwGetCursorPos(window, &mPosX, &mPosY);
   if (mouseDown) {
      camX += (float)(mPosX - mX) / width * M_PI * 2;
      camY -= (float)(mPosY - mY) / height * M_PI;
      camY = camY > yMinMax ? yMinMax : camY;
      camY = camY < -yMinMax ? -yMinMax : camY;
      //printf("Mouse change: (%f, %f)\n", camX, camY);
   }
   Vector3f deltaPos = vec3(0);
   if (wDown) deltaPos(2) += 1;
   if (sDown) deltaPos(2) -= 1;
   if (spaceDown) deltaPos(1) += 1;
   if (shiftDown) deltaPos(1) -= 1;
   if (dDown) deltaPos(0) += 1;
   if (aDown) deltaPos(0) -= 1;

	// Clear framebuffer.
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Use the matrix stack for Lab 6
   aspect = width/(float)height;

   // Create the matrix stacks - please leave these alone for now
   auto M = make_shared<MatrixStack>();
   // Apply perspective projection.
   P->pushMatrix();
   P->perspective(45.0f, aspect, 0.01f, 1000.0f);

   Vector3f lookat = Vector3f(cosf(camY) * cosf(camX), 
                              sinf(camY), 
                              cosf(camY) * cosf(M_PI/2 - camX));
   camRotation = lookat;
   
   if (deltaPos(0) || deltaPos(1) || deltaPos(2)) {
      Vector3f mover = vec3(0);
      deltaPos.normalize();
      lookat.normalize();
      if (deltaPos(0) || deltaPos(2)) {
         float theta = atan2f(deltaPos(2), deltaPos(0)) - atan2f(lookat(2), lookat(0));
         mover = vec3(sinf(theta), 0, cosf(theta));
      }
      if (deltaPos(1)) {
         mover = vec3(mover(0), deltaPos(1), mover(2));
      }
      playerPos += mover * speed;
      if (playerPos(1) < 0) playerPos = vec3(playerPos(0), 0, playerPos(2));
   }
   
   if (mousePressed) {
      tracePosition = traceLine(scene, 0, 200, 1);
   }

   V->pushMatrix();
		V->lookAt(playerPos, lookat + playerPos, Vector3f(0, 1, 0));
		
	
	// Draw a stack of cubes with indiviudal transforms 
   glEnable(GL_DEPTH_TEST);

	M->pushMatrix();
		M->loadIdentity();
		drawGameObject(M, scene);
	M->popMatrix();

	V->popMatrix();
	P->popMatrix();
	
	drawPoints();
	
	mousePressed = false;
	
	if (cleanDead) {
	   for (int i = 0; i < scene->numChildren; i++) {
	      if (scene->children[i].flag & SPHERE) {
	         int removed = 0;
	         for (int j = 0; j + removed < scene->children[i].numChildren; j++) {
	            if (scene->children[i].children[j+removed].flag & DEAD) {
	               ++removed;
	            }
               scene->children[i].children[j] = scene->children[i].children[j+removed];
	         }
	         scene->children[i].numChildren -= removed;
	      }
	   }
	   cleanDead = false;
	}
}

int main(int argc, char **argv) {

	// Set error callback.
	glfwSetErrorCallback(error_callback);
	// Initialize the library.
	if(!glfwInit()) {
		return -1;
	}
	
	srand(time(NULL));
   //request the highest possible version of OGL - important for mac
   glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
   glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
   glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
   glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 2);

	// Create a windowed mode window and its OpenGL context.
	window = glfwCreateWindow(1280, 720, "Kyle Nuzzolo Program 2", NULL, NULL);
	if(!window) {
		glfwTerminate();
		return -1;
	}
	// Make the window's context current.
	glfwMakeContextCurrent(window);
	// Initialize GLEW.
	glewExperimental = true;
	if(glewInit() != GLEW_OK) {
		cerr << "Failed to initialize GLEW" << endl;
		return -1;
	}
	//weird bootstrap of glGetError
   glGetError();
	cout << "OpenGL version: " << glGetString(GL_VERSION) << endl;
   cout << "GLSL version: " << glGetString(GL_SHADING_LANGUAGE_VERSION) << endl;

	// Set vsync.
	glfwSwapInterval(1);
	// Set keyboard callback.
	glfwSetKeyCallback(window, key_callback);
   //set the mouse call back
   glfwSetMouseButtonCallback(window, mouse_callback);
   //set the window resize call back
   glfwSetFramebufferSizeCallback(window, resize_callback);

	// Initialize scene. Note geometry initialized in init now
	init();
	GameObject scene = setupScene();

	// Loop until the user closes the window.
	while(!glfwWindowShouldClose(window)) {
	   ++timer;
	   if (timer % 200 == 0) {
	      Vector3f ran = vec3(rand() % 100 - 50, 0, rand() % 100 - 50);
	      ran.normalize();
	      ran *= 75;
	      ran += vec3(0, 25, 0) + playerPos;
	      scene = addChild(scene, createPrefab1(ran, 6));
	   }
		// Render scene.
		render(&scene);
		// Swap front and back buffers.
		glfwSwapBuffers(window);
		// Poll for and process events.
		glfwPollEvents();
	}
	// Quit program.
	glfwDestroyWindow(window);
	glfwTerminate();
	printf("\n");
	return 0;
}
